#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <libspe2.h>
#include <sys/times.h>
#include "vector.h"

extern spe_program_handle_t spe_code;

void * thread(void * data)
{
	unsigned int entry;
	spe_context_ptr_t ctx;
 	argumento *datos = (argumento *)data;
 	
	/* Inicializacion del contexto */
	ctx = spe_context_create(0,NULL);
	entry = SPE_DEFAULT_ENTRY;
	spe_program_load(ctx,&spe_code);

	/* Ejecuta el contexto sobre un SPE, el hilo se bloquea hasta el fin de la ejecucion */
	spe_context_run(ctx,&entry,0,(void *)datos, NULL,NULL);
	spe_context_destroy(ctx);
	return (void *) 0;
}


int main(int argc, char ** argv)
{
	int i,nbspe;
	struct tms tmsaux;
	clock_t ini,fin;
	argumento datosHilo[MAX_SPE] __attribute__((aligned(16)));
	int operando1[LON_VECTOR] __attribute__((aligned(16)));
	int resultado[LON_VECTOR] __attribute__((aligned(16)));
	
	/*lista de los identificadores de pthreads*/
	pthread_t threads[MAX_SPE];

	if (argc < 2)
	{
		printf("Se requiere al menos un parmetro con el nmero de SPEs a usar.\n");
		return -1;
	}
	
	nbspe = atoi(argv[1]);
	if (nbspe>6)
	{
		printf("Error: numero maximo de SPE disponibles = 6\n");
		return -1;
	}

        
        /* Creacin de los vectores a operar */
        for (i=0;i<LON_VECTOR;i++)
        {
        	operando1[i]=i;
        }
        
        /* Creacin de los argumentos de los hilos */

	// ATENCION: LON_VECTOR debe ser mltiplo de 4, 5 y 6 (multiplo de 120)
 	for(i=0;i<nbspe;i++)
	{
		int trozo; // Los trozos deben ser mltiplos de 16 bytes (enteros multiplos de 4)
		trozo=LON_VECTOR/nbspe;
		datosHilo[i].op1=(unsigned long long)&(operando1[i*trozo]);
		datosHilo[i].res=(unsigned long long)&(resultado[i*trozo]);
		datosHilo[i].lon=trozo;
	}
	
        
        ini=times(&tmsaux);   // Inicio del tiempo de ejecucion
        
        /* Creacion de los hilos */
	for(i=0;i<nbspe;i++)
	{
		/*Iniciar un hilo por cada SPE */
		if (pthread_create(threads+i,NULL,thread,&(datosHilo[i])) != 0)
		{
			perror("Thread create.");
			return -1;
		}
	}
	
	/* Esperar el fin de cada hilo, la funcion pthread_join espera el fin del thread, y recoge el valor devuelto por este en el segundo argumento */
	for(i=0;i<nbspe;i++)
	{
		pthread_join(threads[i],NULL);
	}
	
	
	fin=times(&tmsaux);  // Final del tiempo de ejecucin
	
	/* Muestra resultados */
	
	// comprueba resultado
	for (i=0;i<LON_VECTOR;i++)
	{
	  if (operando1[i]-1!=resultado[i])
	    printf("%3d -> %3d\n",operando1[i],resultado[i]);
	}
	
	printf("\nTiempo: %ld\n",	fin-ini);
	
	printf("\nFin del programa.\n");

	return 0;
}
